#include <cstdio>
#include <cstdarg>
#include <cstring>
#include <cmath>
#include <memory>
#include <iostream>

#include "matrix"
#include "meps.h"

//#define KECY

/* default English error messages */
static char *merrmessages[] = {
  NULL,
  "invalid matrix size",
  "out of bounds",
  "incompatible matrix size",
  "not enough memory",
  "not enough memory for auxiliary operation",
  "block matrix is not initialized",
  "unknown error",
  NULL };
  
using namespace std;

CMatrixError::CMatrixError(int _id)
{
  id = _id;
}

CMatrixError::~CMatrixError() throw()
{
}

int CMatrixError::ID() const
{
  return id;
}

const char* CMatrixError::what() const throw()
{
  if (id <= M_MAX_ERROR)
    return merrmessages[id];
  else
    return merrmessages[M_MAX_ERROR + 1];
}

CMatrixError& CMatrixError::operator = (const CMatrixError& rvalue)
{
  id = rvalue.id;
  return *this;
}

int operator == (const CMatrixError& left, const CMatrixError& right)
{
  return left.ID() == right.ID();
}

CMatrix::CMatrix()
{
#ifdef KECY
  cout << "CMatrix::CMatrix();" << endl;
#endif
  m = n = 0;
  matrix = NULL;
}

CMatrix::CMatrix(int _m, int _n)
{
#ifdef KECY
  cout << "CMatrix::CMatrix(" << _m << ", " << _n << ");" << endl;
#endif
  if ((_m <= 0) || (_n <= 0))
    throw CMatrixError(M_INV_SIZE);
  CreateMatrix(_m, _n);
}

CMatrix::CMatrix(const CMatrix& Matrix)
{
#ifdef KECY
  cout << "CMatrix::CMatrix(Matrix);" << endl;
#endif
  CreateMatrix(Matrix);
}

CMatrix::CMatrix(int _m, int _n, char *_matrix)
{
#ifdef KECY
  cout << "CMatrix::CMatrix(" << _m << ", " << _n << ", \"_matrix\");" << endl;
#endif
  CreateMatrix(_m, _n);
  SetupStr(_matrix);
}

CMatrix::~CMatrix()
{
#ifdef KECY
  cout << "CMatrix::~CMatrix();" << endl;
#endif
  DestroyMatrix();
}

/* Fills the matrix with real numbers
   WARNING: You have to specify the numbers EXACTLY this way:
     Setup(1.0, 2.0, 3.0)
   If you use Setup(1, 2, 3), all numbers except the first one will be
   corrupted. If you need to pass integers, use SetupInt instead. */
void CMatrix::Setup(double d, ...)
{
  va_list arglist;
  int i, j;

  va_start(arglist, d);

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
    {
      matrix[i*n+j] = d;
      d = va_arg(arglist, double);
    }
  
  va_end(arglist);
}

/* Fills the matrix with integers. */
void CMatrix::SetupInt(long int l, ...)
{
  va_list arglist;
  int i, j;

  va_start(arglist, l);

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
    {
      matrix[i*n+j] = (double) l;
      l = va_arg(arglist, long int);
    }
  
  va_end(arglist);
}

/* Fills the matrix using a string. */
void CMatrix::SetupStr(char *_matrix)
{
  char *buffer = strdup(_matrix);
  char *p1, *p2;
  int i = 0;

  p2 = buffer;
  while ((p2 != NULL) && (*p2 != '\0'))
  {
    p1 = p2;
    p2 = strchr(p1, ';');
    if (p2 != NULL)
    {
      *p2 = '\0';
      p2++;
    }
    sscanf(p1, "%lf", &matrix[i]);
    i++;
    
    if (i == m * n) /* check for an overflow */
      break;
  }
  
  free(buffer);
}

/* Gets matrix field m,n */
double CMatrix::GetField(int i, int j) const
{
  if ((i < 0) || (j < 0) || (i >= m) || (j >= n))
    throw CMatrixError(M_OUT_OF_BOUNDS);
  return matrix[i*n+j];
}

/* Sets matrix field m,n */
void CMatrix::SetField(int i, int j, double d)
{
  if ((i < 0) || (j < 0) || (i >= m) || (j >= n))
    throw CMatrixError(M_OUT_OF_BOUNDS);
  matrix[i*n+j] = d;
}

/* Gets matrix dimensions */
void CMatrix::GetSize(int &_m, int &_n) const
{
  _m = m; _n = n;
}
int CMatrix::GetRows()
{
  return m;
}
int CMatrix::GetCols()
{
  return n;
}

/* Gets matrix */
double *CMatrix::GetMatrix()
{
  return matrix;
}
double *CMatrix::GetMatrix(int &_m, int &_n)
{
  _m = m;
  _n = n;
  return matrix;
}

/* return this matrix as transposed */
CMatrix CMatrix::MakeTransposed()
{
  int i, j;
  CMatrix Matrix(n, m);

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      Matrix.SetField(j, i, GetField(i, j));

  return Matrix;
}

/* transpose this matrix */
void CMatrix::Transpose()
{
  CMatrix Matrix;

  Matrix = MakeTransposed();
  
  DestroyMatrix();
  CreateMatrix(Matrix);
}

CMatrix& CMatrix::operator = (const CMatrix& rvalue)
{
  DestroyMatrix();
  CreateMatrix(rvalue);
  return *this;
}

CMatrix& CMatrix::operator += (const CMatrix& rvalue)
{
  int i, j;
  int _m, _n;

  rvalue.GetSize(_m, _n);
  if ((m != _m) || (n != _n))
    throw CMatrixError(M_BAD_SIZE);

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      SetField(i, j, rvalue.GetField(i, j) + GetField(i, j));

  return *this;
}

CMatrix& CMatrix::operator -= (CMatrix& rvalue)
{
  CMatrix Matrix(-rvalue);
  *this += Matrix;
  return *this;
}

CMatrix& CMatrix::operator *= (double rvalue)
{
  int i, j;

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      SetField(i, j, GetField(i, j) * rvalue);

  return *this;
}

CMatrix& CMatrix::operator *= (CMatrix& rvalue)
{
  *this = *this * rvalue;
  return *this;
}

/* matrix addition */
CMatrix operator + (const CMatrix& left, const CMatrix& right)
{
  int i, j;
  CMatrix Matrix;
  int _m, _n;
  int _m2, _n2;

  left.GetSize(_m, _n);
  right.GetSize(_m2, _n2);

  /* incompatible matrices? */
  if ((_m != _m2) || (_n != _n2))
    throw CMatrixError(M_BAD_SIZE);
  
  Matrix = CMatrix(_m, _n);
  for (i = 0; i < _m; i++)
    for (j = 0; j < _n; j++)
    Matrix.SetField(i, j, left.GetField(i, j) + right.GetField(i, j));

  return Matrix;
}

/* matrix multiplication (scalar) */
CMatrix operator * (const CMatrix& left, double right)
{
  int i, j;
  int _m, _n;
  CMatrix Matrix(left);

  Matrix.GetSize(_m, _n);
  for (i = 0; i < _m; i++)
    for (j = 0; j < _n; j++)
      Matrix.SetField(i, j, Matrix.GetField(i, j) * right);

  return Matrix;
}

CMatrix operator * (double left, const CMatrix& right)
{
  /* just invert the order... */
  return right * left;
}

/* matrix negation */
CMatrix operator - (const CMatrix& right)
{
  return right * -1;
}

/* matrix substraction */
CMatrix operator - (const CMatrix& left, const CMatrix& right)
{
  CMatrix Matrix(-right);
  return left + Matrix;
}

/* matrix multiplication */
CMatrix operator * (const CMatrix& left, const CMatrix& right)
{
  int _m1, _n1, _m2, _n2;
  int i, j, k;
  double d;

  left.GetSize(_m1, _n1);
  right.GetSize(_m2, _n2);

  /* incompatible matrices? */
  if (_n1 != _m2)
    throw CMatrixError(M_BAD_SIZE);
  
  CMatrix Matrix(_m1, _n2);
  
  for (i = 0; i < _m1; i++)
    for (j = 0; j < _n2; j++)
    {
      d = 0;
      for (k = 0; k < _n1; k++)
        d += left.GetField(i, k) * right.GetField(k, j);
      Matrix.SetField(i, j, d);
    }

  return Matrix;
}

/* matrix comparison */
int operator == (const CMatrix& left, const CMatrix& right)
{
  int _m, _n;
  int _m2, _n2;
  int i, j;
  
  left.GetSize(_m, _n);
  right.GetSize(_m2, _n2);

  /* incompatible matrices? */
  if ((_m != _m2) || (_n != _n2))
    throw CMatrixError(M_BAD_SIZE);


  for (i = 0; i < _m; i++)
    for (j = 0; j < _n; j++)
    {
//      if (fabs(log(left.GetField(i, j)) - log(right.GetField(i, j))) >= lf_eps2)
//      if (fabs(log10(left.GetField(i, j)) - log10(right.GetField(i, j))) >= lf_eps2)
      if (fabs(left.GetField(i, j) - right.GetField(i, j)) >= 
        lf_eps2 * (lf_eps2 + right.GetField(i, j)))
        return 0;
    }
  return 1;

//  return !memcmp(left.GetMatrix(), right.GetMatrix(), sizeof(double) * _m * _n);
}

int operator != (const CMatrix& left, const CMatrix& right)
{
  return !(left == right);
}

/* Creates a new matrix (m) x (n) (m rows and n columns) */
void CMatrix::CreateMatrix(int _m, int _n)
{
  /* just for safety... */
  if ((_m <= 0) || (_n <= 0))
    throw CMatrixError(M_INV_SIZE);

  m = _m;
  n = _n;

  try {
    matrix = new double[m*n];
    if (matrix == NULL)
      throw CMatrixError(M_NO_MEM);
  } catch(bad_alloc) { throw CMatrixError(M_NO_MEM); }
}

void CMatrix::CreateMatrix(const CMatrix& Matrix)
{
  /* Create uninitialized matrix */
  CreateMatrix(Matrix.m, Matrix.n);

  /* Copy the data */
  memcpy(matrix, Matrix.matrix, Matrix.m * Matrix.n * sizeof(double));
}

void CMatrix::DestroyMatrix()
{
  if (matrix != NULL)
    delete[] matrix;
  matrix = NULL;
  m = n = 0;
}
