#include <cstdio>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <iostream>
#include <fstream>

#define Uses_Matrix
#include "matrix"

#include "meps.h"

using namespace std;

#define USE_FILE 1
#define USE_RAND 2
int action = 0;
char infile[256] = { 0, };
int rand_m = 0, rand_n = 0;

void checkparameters(int argc, char *argv[])
{
  int i;

  for (i = 1; i < argc; i++)
  {
    if (argv[i][0] == '-')
    {
      if (!strcmp(argv[i]+1, "file"))
      {
        if (i + 1 < argc)
        {
          action |= USE_FILE;
          strcpy(infile, argv[++i]);
        }
        else
          cout << "Warning: missing file name" << endl;
      }

      if (!strcmp(argv[i]+1, "rand"));
      {
        if (i + 1 < argc)
        {
          if (sscanf(argv[++i], "%d", &rand_m) == 1)
            action |= USE_RAND;
          else
            cout << "Warning: missing matrix size" << endl;
        }
        
        if (i + 1 < argc)
        {
          if (sscanf(argv[++i], "%d", &rand_n) < 1)
            rand_n = 0;
        }
      }
    }
  }
}

int main(int argc, char *argv[])
{
  FILE *in;
  int m_rows, m_cols;
  int i, j;
  double r;
  clock_t ticks;

  srand(time(0));
  checkparameters(argc, argv);

  if (action == 0)
  {
    printf("Test matrix program.\n"
           "Syntax: matrixtest <-parameter> [input_file]\n"
           "  -file input_file - load data from text file\n"
           "  -rand m [n]      - create random matrix with size m,n\n"
           "                   - if you omit n, m+1 is used instead\n");
           
    return 0;
  }

  try {
    CGaussMatrix Matrix;
    
    printf("Using lf_eps2: %e\n", lf_eps2);

    if (action & USE_FILE) /* file mode */
    {
      cout << "Using file \"" << infile << "\" as an input..." << endl;
      in = fopen(infile, "r");
      if (in == NULL)
      {
        cout << "Unable to open '" << infile << "'" << endl;
        return 1;
      }
  
      fscanf(in, "%d,%d", &m_rows, &m_cols);
  
      cout << "Creating new matrix (" << m_rows << "," << m_cols << ")" << endl;
    
      Matrix = CGaussMatrix(m_rows, m_cols);
    
      for (i = 0; i < m_rows; i++)
        for (j = 0; j < m_cols; j++)
        {
          fscanf(in, "%lf", &r);
          Matrix.SetField(i, j, r);
        }
      fclose(in);
    }
    
    if (action & USE_RAND) /* file mode */
    {
      if (rand_m <= 0)
      {
        cout << "Error: incorrect count of rows: " << rand_m << endl;
        return 1;
      }
      if (rand_n <= 0)
        rand_n = rand_m + 1;
      m_rows = rand_m;
      m_cols = rand_n;

      cout << "Creating random matrix " << m_rows << "x" << m_cols << endl;

      Matrix = CGaussMatrix(m_rows, m_cols);
      for (i = 0; i < m_rows; i++)
        for (j = 0; j < m_cols; j++)
          Matrix.SetField(i, j, (double) rand() / (double) RAND_MAX);
    }

    PrintMatrix(Matrix);
    
    /* matrices other than m,m+1 cannot be usually eliminated properly */
    if (m_rows + 1 < m_cols)
    {
      cout << "Warning - this matrix won't be fully eliminated." << endl;
    }
    
    CGaussMatrix Matrix2 = Matrix;
    cout << "Computing..." << endl;
    ticks = clock();
    Matrix2.LinearEquation();
    ticks = clock() - ticks;
    cout << "Result:" << endl;
    PrintMatrix(Matrix2);

    cout << "Time: " << (float) ticks / CLOCKS_PER_SEC << " seconds" << endl;

    if (m_rows + 1 >= m_cols)
    {
      int stop = 0, limit = m_rows;
      
      cout << "Verify: ";
      
      if (m_rows + 1 != m_cols)
      {
        for (i = m_cols - 1; i < m_rows; i++)
          if (Matrix2.GetField(i, m_cols - 1) >= lf_eps2)
          {
            cout << "no result" << endl;
            stop = 1;
            break;
          }
        limit = m_cols - 1;
      }

      if (!stop)
      {
        /* split matrix into blocks A and B */
        if (limit == m_rows)
        {
          Matrix.SetupBlockMatrix(1, BM_COUNT, m_cols - 1, 1, BM_USERDEFINED);
          Matrix2.SetupBlockMatrix(1, BM_COUNT, m_cols - 1, 1, BM_USERDEFINED);
        }
        else
        {
          Matrix.SetupBlockMatrix(limit, m_rows - limit, BM_USERDEFINED, m_cols - 1, 1, BM_USERDEFINED);
          Matrix2.SetupBlockMatrix(limit, m_rows - limit, BM_USERDEFINED, m_cols - 1, 1, BM_USERDEFINED);
        }

        CGaussMatrix Matrix3 =
          Matrix.GetBlockMatrix(0, 0) * Matrix2.GetBlockMatrix(0, 1);
	  
        cout << (Matrix3 == Matrix.GetBlockMatrix(0, 1) ? "OK" : "failed") << endl;
      }
//    PrintMatrix(Matrix3);
//    PrintMatrix(Matrix.GetBlockMatrix(0, 1));
    }
  } catch (CMatrixError &Error)
  {
    cout << "Matrix error #" << Error.ID() << ": " << Error.what() << endl;
  } catch (exception &Error)
  {
    cout << "UNHANDLED EXCEPTION: " << Error.what() << endl;
  }
}
